@@ -28,8 +28,10 @@
 ;; the mu4e context.
 
 
-;; Code
+;;; Code:
 (require 'message)
+(require 'sendmail)
+(require 'gnus-msg)
 (require 'nnheader) ;; for make-full-mail-header
 
 (require 'mu4e-obsolete)
@@ -77,7 +79,7 @@ A symbol:
 - nil           : default (new buffer)
 - window        : compose in new window
 - frame or t    : compose in new frame
-- display-buffer: use display-buffer / display-buffer-alist
+- display-buffer: use `display-buffer' / `display-buffer-alist'
   (for fine-tuning).
 
 For backward compatibility with `mu4e-compose-in-new-frame', t is
@@ -140,7 +142,7 @@ All `sign-*' options have a `encrypt-*' analogue."
               (const :tag "Encrypt all messages" encrypt-all-messages)
               (const :tag "Sign new messages" sign-new-messages)
               (const :tag "Encrypt new messages" encrypt-new-messages)
-y              (const :tag "Sign forwarded messages" sign-forwarded-messages)
+              (const :tag "Sign forwarded messages" sign-forwarded-messages)
               (const :tag "Encrypt forwarded messages"
                      encrypt-forwarded-messages)
               (const :tag "Sign edited messages" sign-edited-messages)
@@ -192,6 +194,97 @@ messages, it is nil.")
 (defvar-local mu4e-compose-type nil
   "The compose-type for the current message.")
 
+(defvar mu4e-captured-message)
+(defun mu4e-compose-attach-captured-message ()
+  "Insert the last captured message file as an attachment.
+Messages are captured with `mu4e-action-capture-message'."
+  (interactive)
+  (if-let* ((msg mu4e-captured-message)
+            (path (plist-get msg :path))
+            (path (and (file-exists-p path) path)))
+      (mml-attach-file
+       path
+       "message/rfc822"
+       (or (plist-get msg :subject) "No subject")
+       "attachment")
+    (mu4e-warn "No valid message has been captured")))
+
+;; Go to bottom / top
+
+(defun mu4e-compose-goto-top (&optional arg)
+  "Go to the beginning of the message or buffer.
+Go to the beginning of the message or, if already there, go to
+the beginning of the buffer.
+
+Push mark at previous position, unless either a
+\\[universal-argument] prefix ARG is supplied, or Transient Mark mode
+is enabled and the mark is active."
+  (interactive "P")
+  (or arg
+      (region-active-p)
+      (push-mark))
+  (let ((old-position (point)))
+    (message-goto-body)
+    (when (equal (point) old-position)
+      (goto-char (point-min)))))
+
+(defun mu4e-compose-goto-bottom (&optional arg)
+  "Go to the end of the message or buffer.
+Go to the end of the message (before signature) or, if already
+there, go to the end of the buffer.
+
+Push mark at previous position, unless either a
+\\[universal-argument] prefix ARG is supplied, or Transient Mark mode
+is enabled and the mark is active."
+  (interactive "P")
+  (or arg
+      (region-active-p)
+      (push-mark))
+  (let ((old-position (point))
+        (message-position (save-excursion (message-goto-body) (point))))
+    (goto-char (point-max))
+    (when (re-search-backward message-signature-separator message-position t)
+      (forward-line -1))
+    (when (equal (point) old-position)
+      (goto-char (point-max)))))
+
+(defun mu4e-compose-context-switch (&optional force name)
+  "Change the context for the current draft message.
+
+With NAME, switch to the context with NAME, and with FORCE non-nil,
+switch even if the switch is to the same context.
+
+Like `mu4e-context-switch' but with some changes after switching:
+1. Update the From and Organization headers as per the new context
+2. Update the `message-signature' as per the new context.
+
+Unlike some earlier version of this function, does _not_ update
+the draft folder for the messages, as that would require changing
+the file under our feet, which is a bit fragile."
+  (interactive "P")
+
+  (unless (derived-mode-p 'mu4e-compose-mode)
+    (mu4e-error "Only available in mu4e compose buffers"))
+
+  (let ((old-context (mu4e-context-current)))
+    (unless (and name (not force) (eq old-context name))
+      (unless (and (not force)
+                   (eq old-context (mu4e-context-switch nil name)))
+        (save-excursion
+          ;; Change From / Organization if needed.
+          (message-replace-header "Organization"
+                                  (or (message-make-organization) "")
+                                  '("Subject")) ;; keep in same place
+          (message-replace-header "From"
+                                  (or (message-make-from) ""))
+          ;; Update signature.
+          (when (message-goto-signature) ;; delete old signature.
+            (if message-signature-insert-empty-line
+                (forward-line -2) (forward-line -1))
+            (delete-region (point) (point-max)))
+          (when message-signature
+              (save-excursion (message-insert-signature))))))))
+
 
 ;;; Filenames
 (defun mu4e--message-basename()
@@ -271,37 +364,42 @@ sending."
   (propertize "--text follows this line--" 'read-only t 'intangible t)
   "Line used to separate headers from text in messages being composed.")
 
-(defun mu4e--compose-complete-contact (&optional start)
-  "Complete the text at START with a contact.
-Ie. either \"name <email>\" or \"email\")."
-  (interactive)
+(defun mu4e-complete-contact ()
+  "Attempt to complete the text at point with a contact.
+I.e., either \"name <email>\" or \"email\". Return nil if not found.
+
+This function can be used for `completion-at-point-functions', to
+complete addresses. This can be used outside mu4e, but mu4e must
+be active (running) for this to work."
+  (let* ((end (point))
+         (start (save-excursion
+                  (re-search-backward "\\(\\`\\|[\n:,]\\)[ \t]*")
+                  (goto-char (match-end 0))
+                  (point))))
+    (list start end #'mu4e--compose-complete-handler)))
+
+(defun mu4e--compose-complete-contact-field ()
+  "Attempt to complete a contact when in a contact field.
+
+This is like `mu4e-compose-complete-contact', but limited to the
+contact fields."
   (let ((mail-abbrev-mode-regexp
          "^\\(To\\|B?Cc\\|Reply-To\\|From\\|Sender\\):")
-        (mail-header-separator mu4e--header-separator)
-        (eoh ;; end-of-headers
-         (save-excursion
-           (goto-char (point-min))
-           (search-forward-regexp mail-header-separator nil t))))
-    ;; try to complete only when we're in the headers area, looking at an
-    ;; address field.
-    (when (and eoh (> eoh (point)) (mail-abbrev-in-expansion-header-p))
-      (let* ((end (point))
-             (start
-              (or start
-                  (save-excursion
-                    (re-search-backward "\\(\\`\\|[\n:,]\\)[ \t]*")
-                    (goto-char (match-end 0))
-                    (point)))))
-        (list start end 'mu4e--compose-complete-handler)))))
+        (mail-header-separator mu4e--header-separator))
+    (when (mail-abbrev-in-expansion-header-p)
+      (mu4e-complete-contact))))
 
 (defun mu4e--compose-setup-completion ()
-  "Set up auto-completion of addresses."
-  (set (make-local-variable 'completion-ignore-case) t)
-  (set (make-local-variable 'completion-cycle-threshold) 7)
-  (add-to-list (make-local-variable 'completion-styles) 'substring)
-  (add-hook 'completion-at-point-functions
-            'mu4e--compose-complete-contact nil t))
-
+  "Set up auto-completion of addresses if enabled."
+  ;; turn off message-mode's completion, it's just interfering.
+  (remove-hook 'completion-at-point-functions
+               #'message-completion-function 'local)
+  (when mu4e-compose-complete-addresses
+    (set (make-local-variable 'completion-ignore-case) t)
+    (set (make-local-variable 'completion-cycle-threshold) 7)
+    (add-to-list (make-local-variable 'completion-styles) 'substring)
+    (add-hook 'completion-at-point-functions
+              #'mu4e--compose-complete-contact-field nil t)))
 
 (defun mu4e--fcc-handler (msgpath)
   "Handle Fcc: for MSGPATH.
@@ -347,7 +445,10 @@ If MSGPATH is nil, do nothing."
       (message-narrow-to-headers)
       (unless (message-fetch-field "Message-ID")
         (message-generate-headers '(Message-ID)))
-      (message-generate-headers '(Date)))
+        ;; older Emacsen (<= 28 perhaps?) won't update the Date
+        ;; if there already is one; so make sure it's gone.
+      (message-remove-header "Date")
+      (message-generate-headers '(Date Subject From)))
     (mu4e--delimit-headers 'undelimit))) ;; remove separator
 
 (defvar mu4e--compose-buffer-max-name-length 48)
@@ -386,25 +487,26 @@ message buffer."
   (let ((buf (find-file-noselect path)))
     (when buf
       (with-current-buffer buf
-        (message-narrow-to-headers-or-head)
-        (let ((in-reply-to (message-fetch-field "in-reply-to"))
-              (forwarded-from)
-              (references (message-fetch-field "references")))
-          (unless in-reply-to
-            (when references
-              (with-temp-buffer ;; inspired by `message-shorten-references'.
-                (insert references)
-                (goto-char (point-min))
-                (let ((refs))
-                  (while (re-search-forward "<[^ <]+@[^ <]+>" nil t)
-                    (push (match-string 0) refs))
-                  ;; the last will be the first
-                  (setq forwarded-from (car refs))))))
-          ;; remove the <> and update the flags on the server-side.
-          (when (and in-reply-to (string-match "<\\(.*\\)>" in-reply-to))
-            (mu4e--server-move (match-string 1 in-reply-to) nil "+R-N"))
-          (when (and forwarded-from (string-match "<\\(.*\\)>" forwarded-from))
-            (mu4e--server-move (match-string 1 forwarded-from) nil "+P-N")))))))
+        (save-restriction
+          (message-narrow-to-headers)
+          (let ((in-reply-to (message-fetch-field "in-reply-to"))
+                (forwarded-from)
+                (references (message-fetch-field "references")))
+            (unless in-reply-to
+              (when references
+                (with-temp-buffer ;; inspired by `message-shorten-references'.
+                  (insert references)
+                  (goto-char (point-min))
+                  (let ((refs))
+                    (while (re-search-forward "<[^ <]+@[^ <]+>" nil t)
+                      (push (match-string 0) refs))
+                    ;; the last will be the first
+                    (setq forwarded-from (car refs))))))
+            ;; remove the <> and update the flags on the server-side.
+            (when (and in-reply-to (string-match "<\\(.*\\)>" in-reply-to))
+              (mu4e--server-move (match-string 1 in-reply-to) nil "+R-N"))
+            (when (and forwarded-from (string-match "<\\(.*\\)>" forwarded-from))
+              (mu4e--server-move (match-string 1 forwarded-from) nil "+P-N"))))))))
 
 (defun mu4e--compose-after-save()
   "Function called immediately after the draft buffer is saved."
@@ -441,11 +543,13 @@ appropriate flag at the message forwarded or replied-to."
   ;; Remove References: if In-Reply-To: is missing.
   ;; This allows the user to effectively start a new message-thread by
   ;; removing the In-Reply-To header.
-  (when (eq mu4e-compose-type 'reply)
-    (unless (message-fetch-field "In-Reply-To")
-      (message-remove-header "References")))
-  (when use-hard-newlines
-    (mu4e--send-harden-newlines))
+  (save-restriction
+    (message-narrow-to-headers)
+    (when (eq mu4e-compose-type 'reply)
+      (unless (message-fetch-field "In-Reply-To")
+        (message-remove-header "References")))
+    (when use-hard-newlines
+      (mu4e--send-harden-newlines)))
   ;; for safety, always save the draft before sending
   (set-buffer-modified-p t)
   (save-buffer))
@@ -514,8 +618,18 @@ buffers; lets remap its faces so it uses the ones for mu4e."
 (defvar mu4e-compose-mode-map
   (let ((map (make-sparse-keymap)))
     (set-keymap-parent map message-mode-map)
-    (define-key map (kbd "C-S-u")               #'mu4e-update-mail-and-index)
-    (define-key map (kbd "C-c C-u")             #'mu4e-update-mail-and-index)
+    (define-key map (kbd "C-S-u")    #'mu4e-update-mail-and-index)
+    (define-key map (kbd "C-c C-u")  #'mu4e-update-mail-and-index)
+    (define-key map (kbd "C-c ;")    #'mu4e-compose-context-switch)
+
+    ;; emacs 29
+    ;;(keymap-set map "<remap> <beginning-of-buffer>" #'mu4e-compose-goto-top)
+    ;;(keymap-set map "<remap> <end-of-buffer>" #'mu4e-compose-goto-bottom)
+    (define-key map (vector 'remap #'beginning-of-buffer)
+                #'mu4e-compose-goto-top)
+    (define-key map (vector 'remap #'end-of-buffer)
+                #'mu4e-compose-goto-bottom)
+
     ;; remove some unsupported commands... [remap ..]  does not work here
     ;; XXX remove from menu, too.
     (define-key map (kbd "C-c C-f C-n")   nil) ;; message-goto-newsgroups
@@ -550,11 +664,8 @@ buffers; lets remap its faces so it uses the ones for mu4e."
     (set (make-local-variable 'message-send-mail-real-function) nil)
     ;; Set to nil to enable `electric-quote-local-mode' to work:
     (set (make-local-variable 'comment-use-syntax) nil)
-    ;; offer completion for e-mail addresses
-    (when mu4e-compose-complete-addresses
-      (mu4e--compose-setup-completion))
-    ;; format-flowed
-    (if mu4e-compose-format-flowed
+    (mu4e--compose-setup-completion) ;; maybe offer address completion
+    (if mu4e-compose-format-flowed   ;; format-flowed
         (progn
           (turn-off-auto-fill)
           (setq truncate-lines nil
@@ -596,15 +707,17 @@ With HEADERS-ONLY non-nil, only include the headers part."
     ;; in rare (broken) case, if a message-id is missing use the generated one
     ;; from mu.
     (mu4e--delimit-headers)
-    (unless (message-fetch-field "Message-Id")
-      (goto-char (point-min))
-      (insert (format "Message-Id: <%s>\n" (plist-get msg :message-id))))
+    (save-restriction
+      (message-narrow-to-headers)
+      (unless (message-fetch-field "Message-Id")
+        (goto-char (point-min))
+        (insert (format "Message-Id: <%s>\n" (plist-get msg :message-id)))))
     (mu4e--delimit-headers 'undelimit)
     (ignore-errors (run-hooks 'gnus-article-decode-hook))
     (buffer-substring-no-properties (point-min) (point-max))))
 
 (defun mu4e--compose-cite (msg)
-  "Return a cited version of the ORIG message (a string).
+  "Return a cited version of the ORIG message MSG (a string).
 This function uses `message-cite-function', and its settings apply."
   (with-temp-buffer
     (insert (mu4e-view-message-text msg))
@@ -633,16 +746,20 @@ Based on the value of `mu4e-compose-switch'."
     (_             (mu4e-error "Invalid mu4e-compose-switch"))))
 
 (defun mu4e--fake-pop-to-buffer (name &optional _switch)
-  "A fake `message-pop-to-buffer' which creates NAME.
+  "A fake `message-pop-to-buffer' for creating buffer NAME.
 This is a little glue to use `message-reply', `message-forward'
 etc. We cannot use the normal `message-pop-to-buffer' since we're
 not ready yet to show the buffer in mu4e."
-  (set-buffer (get-buffer-create name))
-  (erase-buffer)
-  (current-buffer))
+  ;; note: we're in a _different_ buffer here, so we need to copy
+  ;; message-reply-header's buffer-local value.
+  (let ((reply-headers message-reply-headers))
+    (set-buffer (get-buffer-create name))
+    (setq-local message-reply-headers reply-headers)
+    (erase-buffer)
+    (current-buffer)))
 
 (defun mu4e--headers (compose-type)
-  "Determine headers needed for message."
+  "Determine headers needed for message based on COMPOSE-TYPE."
   (seq-filter #'identity ;; ensure needed headers are generated.
        `(From Subject Date Message-ID
         ,(when (memq compose-type '(reply forward)) 'References)
@@ -651,7 +768,9 @@ not ready yet to show the buffer in mu4e."
         ,(when message-user-organization 'Organization))))
 
 (defun mu4e--compose-setup-buffer (compose-type compose-func parent)
-  "Set up a buffer for message composition before mu4e-compose-mode.
+  "Set up a buffer for message composition before `mu4e-compose-mode'.
+
+COMPOSE-TYPE is the type of message to creat.
 
 COMPOSE-FUNC is a function / lambda to create the specific type
 of message; it should return (but not show) the created buffer.
@@ -690,18 +809,25 @@ This is mu4e's version of `message-hidden-headers'.")
 
 (defun mu4e--message-is-yours-p (func &rest args)
   "Mu4e advice for `message-is-yours'.
+FUNC is the original function, and ARGS are its arguments.
 Is this address yours?"
   (if (mu4e-running-p)
-      (let ((sender (message-fetch-field "from"))
-            (from (message-fetch-field "sender")))
-        (or (and sender (mu4e-personal-or-alternative-address-p
-                         (car (mail-header-parse-address sender))))
-            (and from (mu4e-personal-or-alternative-address-p
-                       (car (mail-header-parse-address from))))))
+      (save-restriction
+        (message-narrow-to-headers)
+        (let ((sender (message-fetch-field "from"))
+              (from (message-fetch-field "sender")))
+          (or (and sender (mu4e-personal-or-alternative-address-p
+                           (car (mail-header-parse-address sender))))
+              (and from (mu4e-personal-or-alternative-address-p
+                         (car (mail-header-parse-address from)))))))
     (apply func args)))
 
 (defun mu4e--compose-setup-post (compose-type &optional parent)
-  "Prepare the new message buffer."
+  "Prepare the new message buffer.
+
+COMPOSE-TYPE determines the type of message to create. PARENT
+refers to the optional message to start from, i.e., the message
+replied to or forwarded, etc."
   (mu4e-compose-mode)
   ;; remember some variables, e.g for user hooks.
   (setq-local
@@ -731,7 +857,9 @@ Is this address yours?"
       (message-goto-to)
     (if (not (message-field-value "Subject"))
         (message-goto-subject)
-      (message-goto-body)))
+      (pcase message-cite-reply-position
+        ((or 'above 'traditional) (message-goto-body))
+        (_ (when (message-goto-signature) (forward-line -2))))))
   ;; buffer is not user-modified yet
   (set-buffer-modified-p nil)
   (undo-boundary))
@@ -747,7 +875,7 @@ COMPOSE-FUNC is a function / lambda to create the specific type
 of message.
 
 Optionally, SWITCH determines how to find a buffer for the message
-(see SWITCH-FUNCTION in `compose-mail').
+\(see SWITCH-FUNCTION in `compose-mail').
 
 Returns the new buffer."
   (cl-assert (member compose-type '(reply forward edit new)))
@@ -785,11 +913,11 @@ Returns the new buffer."
 (defun mu4e-compose-new (&optional to subject other-headers continue
                                    switch-function yank-action send-actions
                                    return-action &rest _)
-  "This is mu4e's implementation of `compose-mail'. TO, SUBJECT,
-OTHER-HEADERS, CONTINUE, SWITCH-FUNCTION, YANK-ACTION
-SEND-ACTIONS RETURN-ACTION are as described in `compose-mail',
-and to the extend that they do not conflict with mu4e inner
-workings."
+  "Mu4e's implementation of `compose-mail'.
+TO, SUBJECT, OTHER-HEADERS, CONTINUE, SWITCH-FUNCTION,
+YANK-ACTION SEND-ACTIONS RETURN-ACTION are as described in
+`compose-mail', and to the extend that they do not conflict with
+mu4e inner workings."
   (interactive)
   (mu4e--compose-setup
    'new (lambda (_parent)
@@ -849,15 +977,17 @@ must be from current user, as determined through
        (mu4e--delimit-headers)
        ;; message-forward expects message-reply-headers to be set up; here we
        ;; only need message-id & references, rest is for completeness.
-       (setq-local message-reply-headers
-             (make-full-mail-header
-              0
-              (or (message-fetch-field "subject") "none")
-              (or (message-fetch-field "from") "nobody")
-              (message-fetch-field "date")
-              (message-fetch-field "message-id" t)
-              (message-fetch-field "references")
-              0 0 ""))
+       (save-restriction
+         (message-narrow-to-headers)
+         (setq-local message-reply-headers
+                     (make-full-mail-header
+                      0
+                      (or (message-fetch-field "subject") "none")
+                      (or (message-fetch-field "from") "nobody")
+                      (message-fetch-field "date")
+                      (message-fetch-field "message-id" t)
+                      (message-fetch-field "references")
+                      0 0 "")))
        (mu4e--delimit-headers 'undelimit)
        (set-buffer-modified-p nil)
        (message-forward)))))
@@ -877,7 +1007,7 @@ must be from current user, as determined through
 
 ;;;###autoload
 (defun mu4e-compose-resend (address)
-  "Re-send the message at point.
+  "Re-send the message at point to ADDRESS.
 The message is resent as-is, without any editing."
   (interactive
    (list (completing-read
