@@ -640,8 +640,7 @@ As a side-effect, a message that is being viewed loses its
         ;; problems later (#2260, #2508), so let's remove those
         (article-remove-cr)
         (setq-local mu4e--view-message msg)
-        (mu4e--view-render-buffer msg)
-        (setq-local mu4e--view-mime-part-cached nil))
+        (mu4e--view-render-buffer msg))
       (mu4e-loading-mode 0)))
   (unless (mu4e--view-detached-p gnus-article-buffer)
     (with-current-buffer mu4e-linked-headers-buffer
@@ -658,7 +657,9 @@ As a side-effect, a message that is being viewed loses its
       (select-window mu4e~headers-view-win)))
   (with-current-buffer gnus-article-buffer
     (let ((inhibit-read-only t))
-      (run-hooks 'mu4e-view-rendered-hook))))
+      (run-hooks 'mu4e-view-rendered-hook))
+    ;; only needed on some setups; #2683
+    (goto-char (point-min))))
 
 (defun mu4e-view-message-text (msg)
   "Return the rendered MSG as a string."
@@ -708,7 +709,6 @@ determine which browser function to use."
 
 (defun mu4e--view-render-buffer (msg)
   "Render current buffer with MSG using Gnus' article mode."
-  (setq gnus-summary-buffer (get-buffer-create " *appease-gnus*"))
   (let* ((inhibit-read-only t)
          (max-specpdl-size mu4e-view-max-specpdl-size)
          (mm-decrypt-option 'known)
